#!python
# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# The implementation of the Komodo Find and Replace service

import os
import sys
import string, types
import re
import logging
import threading

from xpcom import components, nsError, ServerException, COMException
from xpcom.server import UnwrapObject
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
import findlib



#---- globals

log = logging.getLogger("find")
#log.setLevel(logging.DEBUG)

# find types
FOT_SIMPLE = components.interfaces.koIFindOptions.FOT_SIMPLE
FOT_WILDCARD = components.interfaces.koIFindOptions.FOT_WILDCARD
FOT_REGEX_PYTHON = components.interfaces.koIFindOptions.FOT_REGEX_PYTHON

# case sensitivity options
FOC_INSENSITIVE = components.interfaces.koIFindOptions.FOC_INSENSITIVE
FOC_SENSITIVE = components.interfaces.koIFindOptions.FOC_SENSITIVE
FOC_SMART = components.interfaces.koIFindOptions.FOC_SMART

# services setup in Find Service ctor
gPrefSvc = None
lastErrorSvc = None


#---- Find in Files backend

class _UserAbort(Exception):
    """Too facilitate stopping processing when user aborts asynchronous
    search.
    """
    pass

class _FinderInFiles(threading.Thread):
    """Find all hits of the given pattern in the given files and
    load them into the given koIFindResultsView.
    
    The running thread can be terminated early by calling the .stop() method.

    XXX:TODO:
    - Need proper lock-synchronization around usages of self._stop, for
      example.
    """
    def __init__(self, id, pattern, patternType, caseSensitivity, matchWord,
                 folders, cwd, searchInSubfolders, includeFiletypes,
                 excludeFiletypes, resultsMgr, resultsView):
        threading.Thread.__init__(self)

        self.id = id
        self.pattern = pattern
        self.patternType = patternType
        self.caseSensitivity = caseSensitivity
        self.matchWord = matchWord
        self.folders = folders
        self.cwd = cwd
        self.searchInSubfolders = searchInSubfolders
        self.includeFiletypes = includeFiletypes
        self.excludeFiletypes = excludeFiletypes
        self.resultsMgr = resultsMgr
        self.resultsMgrProxy = getProxyForObject(None,
            components.interfaces.koIFindResultsTabManager,
            self.resultsMgr, PROXY_ALWAYS | PROXY_SYNC)
        self.resultsView = resultsView
        self.resultsViewProxy = getProxyForObject(None,
            components.interfaces.koIFindResultsView,
            self.resultsView, PROXY_ALWAYS | PROXY_SYNC)
        self._resultsView = UnwrapObject(resultsView)
        
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                      .getService(components.interfaces.koILastErrorService)

        self._stop = 0 # when true the processing thread should terminate

    def stop(self):
        """Stop processing."""
        log.debug("stopping find in files thread")
        self._stop = 1

    def run(self):
        # Rule: if self._stop is true then this code MUST NOT use
        #       self.resultsMgrProxy or self.resultsViewProxy, because they
        #       may have been destroyed.
        log.debug("start find in files thread")

        # optimization: delay reporting of results, store them here
        self._resetResultCache()

        self.numResults = 0
        self.numFiles = 0
        try:
            #---- Phase 1: get a list of files to search.
            # The plan: For each folder to search get one big string that
            # will have one file per line. This list of files will be
            # *all* files under the given folder. With this string we
            # can use a regex built from the 'includeFiletypes' to
            # efficiently pick out the files to actually search. Similarly
            # for 'excludeFiletypes'.
            # 
            # Notes:
            # - The use of regex searching for '*Filetypes' selection
            #   means that we can expose a regex filetypes syntax to the
            #   user (as does Agent Ransack) if we so choose in the
            #   future.
            # - Eventually phase 1 can be optimized to cache its filelists
            #   for a Komodo session and use OS/FS facilities to track
            #   directory changes.
            # - XXX What about skipping some files? E.g. CVS control files.
            # - XXX Watch symlinks.
            # - XXX How to deal with relative paths? This must be handled
            #       by the caller of koIFindService.findallinfiles().
            if not self._stop:
                self.resultsMgrProxy.setDescription(
                    "Phase 1: gathering list of files...", 0)
            filelistA = ""  # all files in the given folders (one file per line)
            for folder in self.folders:
                if not sys.platform.startswith("win"):
                    folder = os.path.expanduser(folder)
                if not os.path.isabs(folder) and self.cwd is not None:
                    folder = os.path.join(self.cwd, folder)
                    #if not self._stop:
                    #    self.resultsMgrProxy.setDescription(
                    #        "ignored relative path: '%s'" % folder, 1)
                    #log.warn("skipping relative folder '%s', all folders "
                    #         "must be absolute", folder)
                log.debug("generating file list for folder: '%s'", folder)
                oneFileList = ""
                try:
                    oneFileList = self._getFilelist(folder,
                                                    self.searchInSubfolders)
                except EnvironmentError, ex:
                    log.warn("skipping '%s': %s", folder, ex)
                else:
                    if oneFileList:
                        #print "\n---- Files under '%s':\n%s\n\n"\
                        #      % (folder, oneFileList)
                        filelistA += "\n" + oneFileList
                if self._stop:
                    return

            # Convert the list of "wildcard"-type filetypes into a single
            # appropriate regex.
            iftPatterns = [re.escape(f)     # 'ift' == Include FileTypes
                             .replace("\\?", ".")
                             .replace("\\*", ".*")
                           for f in self.includeFiletypes if f]
            if not iftPatterns:
                iftPattern = "^(.*)$"  # empty include filetypes mean match all files
            else:
                # The user's specified filetype patterns only applies to the
                # file's basename. In the following pattern the stuff in
                # the parentheses is the file basename.
                iftPattern = "^.*%s(%s)$" % (re.escape(os.sep),
                                             "|".join(iftPatterns))
            log.debug("include filetypes %r -> regex %r",
                      self.includeFiletypes, iftPattern)
            eftPatterns = [re.escape(f)     # 'eft' == Exclude FileTypes
                             .replace("\\?", ".")
                             .replace("\\*", ".*")
                           for f in self.excludeFiletypes if f]
            if not eftPatterns:
                eftPattern = "^$"  # empty exclude filetypes mean match no files
            else:
                # The user's specified filetype patterns only applies to the
                # file's basename. In the following pattern the stuff in
                # the parentheses is the file basename.
                eftPattern = "^.*%s(%s)$" % (re.escape(os.sep),
                                             "|".join(eftPatterns))
            log.debug("exclude filetypes %r -> regex %r",
                      self.excludeFiletypes, eftPattern)

            # Pick out files matching the user's (filetype) criteria.
            log.debug("filter filelists by filetype: include %r, exclude %r",
                      self.includeFiletypes, self.excludeFiletypes)
            if sys.platform.startswith("win"):
                case = "insensitive"
            else:
                case = "sensitive"
            filesB = [] # included files from filelistA
            matches = findlib.findallex(filelistA, iftPattern,
                                        patternType="regex-python",
                                        case=case)
            for match in matches:
                file = filelistA[match.start():match.end()]
                if file: # skip empty string matches
                    #print "match %d-%d: '%s'"\
                    #      % (match.start(), match.end(), file)
                    filesB.append(file)
            filelistB = '\n'.join(filesB)
           
            matches = findlib.findallex(filelistB, eftPattern,
                                        patternType="regex-python",
                                        case=case)
            if matches:
                filelistC = ""
                index = 0
                for match in matches:
                    filelistC += filelistB[index:match.start()]
                    index = match.end()
                filelistC += filelistB[index:]
            else:
                filelistC = filelistB
            files = [f for f in filelistC.splitlines(0) if f]
            #import pprint
            #pprint.pprint(files)

            #---- Phase 2: searching each file
            searchedFiles = {} # list of files searched, to avoid dupes
            for file in files:
                if self._stop:
                    return
                if file in searchedFiles:
                    continue
                searchedFiles[file] = 1
                try:
                    self._searchFile(file)
                except COMException, ex:
                    errcode, errmsg = self.lastErrorSvc.getLastError()
                    log.warn("skipping '%s': %s: %s", file, errcode, errmsg)
                except EnvironmentError, ex:
                    log.warn("skipping '%s': %s", file, ex)
            self._flushResultCache() # Add the last chunk of find results
        finally:
            if not self._stop:
                self.resultsMgrProxy.searchFinished(1, self.numResults,
                                                    self.numFiles)
                if not files:
                    self.resultsMgrProxy.setDescription(
                        "No files were found to search in.", 1)

        log.debug("done find in files thread")

    def _searchFile(self, file):
        """Search the given file and report results.
        
            "file" is the filename to search.
            
        Search results are reported to self.resultsMgrProxy,
        self.resultsViewProxy and self.numResults.
        """
        #log.debug("search '%s'", file)
        
        # Read the file.
        #XXX Might want to do unicode-y, Komodo preference-y stuff here at
        #    some point.
        fin = open(file, 'rb')
        content = fin.read()
        fin.close()

        # Find any matches.
        matches = findlib.findallex(content, self.pattern,
                                    patternType=self.patternType,
                                    case=self.caseSensitivity,
                                    matchWord=self.matchWord)

        # Report results.
        if matches:
            # Determine line offsets for finding the line and column number
            # of each match.
            lines = [] # One for each line: (start offset, line content)
            offset = 0
            for line in content.splitlines(1):
                lines.append( (offset, line) )
                offset += len(line)

            lower = 0 # initialize out here because matches are sequential
            for match in matches:
                start = match.start()
                end = match.end()
                value = content[start:end]
                #print "\tmatch: %d-%d: %r" % (start, end, value)
                # Determine the line and line offset of this match.
                # - Algorithm adapted from scintilla's LineFromPosition.
                if not lines:
                    lineNum = 0  # lineNum is 0-based
                    lineOffset, line = 0, ""
                elif start >= lines[-1][0]:
                    lineNum = len(lines)
                    lineOffset, line = lines[-1]
                else:
                    upper = len(lines)
                    while lower < upper:
                        middle = (upper + lower + 1) / 2
                        if start < lines[middle][0]:
                            upper = middle - 1
                        else:
                            lower = middle
                    lineNum = lower
                    lineOffset, line = lines[lower]
                columnNum = start - lineOffset

                file = os.path.normpath(file)

                # Optimization: cache find results to only call through
                # XPCOM to log them every 25 or so (see AddFindResults()).
                #self.resultsViewProxy.AddFindResult(
                #    file, start, end, value, file,
                #    lineNum+1, columnNum+1,
                #    line)
                
                # when line gets passed to AddFindResults pyxpcom will convert it into
                # a unicode string automagicaly.  However, if we happened to get a
                # result on a binary file, that will fail and we will not see any results.
                # protect against that by trying to make the string unicode, and failing
                # that, repr it, then limit it to 100 bytes as it could be *very* long.
                if type(line) != types.UnicodeType:
                    try:
                        line = unicode(line)
                    except UnicodeError, e:
                        line = line[:100]
                        line = repr(line)[1:-1]
                
                self._cacheResult(file, start, end, value, file,
                                  lineNum+1, columnNum+1, line.rstrip())

                self.numResults += 1
            self.numFiles += 1

            if self.numResults % 25 == 0:
                self._flushResultCache()

    def _resetResultCache(self):
        self._r_urls = []
        self._r_startIndexes = []
        self._r_endIndexes = []
        self._r_values = []
        self._r_fileNames = []
        self._r_lineNums = []
        self._r_columnNums = []
        self._r_contexts = []

    def _cacheResult(self, url, startIndex, endIndex, value, fileName,
                     lineNum, columnNum, context):
        self._r_urls.append(url)
        self._r_startIndexes.append(startIndex)
        self._r_endIndexes.append(endIndex)
        self._r_values.append(value)
        self._r_fileNames.append(fileName)
        self._r_lineNums.append(lineNum)
        self._r_columnNums.append(columnNum)
        self._r_contexts.append(context)

    def _flushResultCache(self):
        if self._r_urls:
            self.resultsViewProxy.AddFindResults(
                self._r_urls, self._r_startIndexes, self._r_endIndexes,
                self._r_values, self._r_fileNames,
                self._r_lineNums, self._r_columnNums, self._r_contexts)
            self._resetResultCache()

            desc = "Found %d occurrences in %d file(s) so far."\
                   % (self.numResults, self.numFiles)
            self.resultsMgrProxy.setDescription(desc, 0)


    def _getFilelist(self, folder, searchInSubfolders):
        """Return a "list" of files in the given folder.
        
            "folder" is the folder to search.
            "searchInSubFolders" is a boolean indicating whether to recurse
                into the folder.
        
        The "list" of files is returned as a string with one line for
        each file. (Limitation: cannot handle filenames with newlines, but
        that would be crazy.) Returns the empty string if no files are
        found. May raise an EnvironmentError -- it does not guard against,
        for example, 'folder' not existing.
        """
        log.debug("_getFilelist(folder='%s', searchInSubfolders=%s)", folder,
                  searchInSubfolders)
        if not searchInSubfolders:
            filesAndDirs = os.listdir(folder)
            filelist = []
            for f in filesAndDirs:
                fileOrDir = os.path.join(folder, f)
                if os.path.isfile(fileOrDir):
                    filelist.append(fileOrDir)
        else:
            filelist = []
            try:
                os.path.walk(folder, self._addFilesWalker, filelist)
            except _UserAbort:
                pass
        
        return "\n".join(filelist)
    
    def _addFilesWalker(self, filelist, dirname, names):
        """Add the _files_ in this directory to filelist."""
        if self._stop:
            raise _UserAbort("aborted in dir '%s'" % dirname)
        for name in names:
            fullpath = os.path.join(dirname, name)
            if os.path.isfile(fullpath):
                filelist.append(fullpath)

    #def _readFile(self, file):
    #    """Return a Unicode object of the file's content.
    #    
    #        "file" is the name of the file to open and read.
    #        XXX For local PythoAdd some preferences...
    #        
    #    It would be nice to have a Python-only implementation of this,
    #    but for now we will use Komodo-specific code. For a Python-only
    #    impl we would need to provide a mechanism for providing some
    #    preferences/directives for detection handling similar to Komodo's.
    #    
    #    Returns a 3-tuple: (ucontent, encoding, bom)
    #    where,
    #        "ucontent" is a unicode buffer of the file's content
    #        "encoding" is a Python encoding name
    #        "bom" is a boolean indicating if the document uses a byte-order
    #            marker
    #    May raise an EnvironmentError if the file cannot be accessed. May
    #    also raise an XPCOM ServerException, in which case the last error
    #    will be set on koILastErrorService.
    #    """
    #    import uriparse
    #    uri = uriparse.localPathToURI(file)
    #    print "XXX %s:" % file
    #    print "\turi: %s" % uri
    #
    #    #docStateMRU = self._globalPrefSvc.getPrefs("docStateMRU");
    #    #if not self.isUntitled and docStateMRU.hasPref(self.file.URI):
    #    #    url = self.file.URI
    #    #    docState = docStateMRU.getPref(url)
    #    #    self.prefs = docState
    #    #else:
    #    #    self.prefs = components.classes['@activestate.com/koPreferenceSet;1'].\
    #    #                             createInstance(components.interfaces.koIPreferenceSet)
    #    #    if self.isUntitled:
    #    #        self.prefs.id = self._untitledName
    #    #    else:
    #    #        self.prefs.id = self.file.URI
    #    #    docStateMRU.setPref(self.prefs)
    #    #self.prefs.parent = self._globalPrefs
    #
    #    encoding_name = self._getStringPref('encodingDefault')
    #    defaultEncoding = self.encodingServices.get_encoding_info(encoding_name)\
    #                   .python_encoding_name
    #    tryEncoding = self._getStringPref('encoding')
    #    tryXMLDecl = self._getBooleanPref('encodingXMLDec')
    #    tryHTMLMeta = self._getBooleanPref('encodingHTML')
    #    tryModeline = self._getBooleanPref('encodingModeline')
    #    #autodetect = self._getBooleanPref('encodingAutoDetect')
    #
    #    import koUnicodeEncoding
    #    ucontent, encoding, bom = koUnicodeEncoding.autoDetectEncoding(
    #        content, tryXMLDecl, tryHTMLMeta, tryModeline, wantThisEncoding,
    #        defaultEncoding)
    #    bom = len(bom) > 0
    #
    #    
    #    doc = self.docSvcProxy.createDocumentFromURI(uri)
    #    enc_info = doc.myDetectEncoding()
    #    print "\t%r, %r, %r"\
    #          % (enc_info[0][:10]+'...'+enc_info[0][-10:],
    #             enc_info[1], enc_info[2])
    #    return enc_info




#---- find and replace components

class KoFindResult:
    _com_interfaces_ = components.interfaces.koIFindResult
    _reg_desc_ = "Find Result"
    _reg_clsid_ = "{0D889F34-0369-4363-BAAB-7465D14EE421}"
    _reg_contractid_ = "@activestate.com/koFindResult;1"

    def __init__(self, url, start, end, value):
        self.url = url
        self.start = start
        self.end = end
        self.value = value


class KoReplaceResult:
    _com_interfaces_ = [components.interfaces.koIFindResult,
                        components.interfaces.koIReplaceResult]
    _reg_desc_ = "Replace Result"
    _reg_clsid_ = "{53056470-F2F0-4a05-B8FF-36DE117C9741}"
    _reg_contractid_ = "@activestate.com/koReplaceResult;1"

    def __init__(self, url, start, end, value, replacement):
        self.url = url
        self.start = start
        self.end = end
        self.value = value
        self.replacement = replacement


class KoFindOptions:
    _com_interfaces_ = components.interfaces.koIFindOptions
    _reg_desc_ = "Find and Replace Session Options"
    _reg_clsid_ = "{3C4AC462-638E-4fa5-A264-3540FEFA558D}"
    _reg_contractid_ = "@activestate.com/koFindOptions;1"

    def __init__(self):
        self.patternTypePrefName = "find-patternType"
        self.caseSensitivityPrefName = "find-caseSensitivity"
        self.matchWordPrefName = "find-matchWord"
        self.searchBackwardPrefName = "find-searchBackward"
        self.preferredContextTypePrefName = "find-preferredContextType"
        self.displayInFindResults2PrefName = "find-displayInFindResults2"
        self.foldersPrefName = "find-folders"
        self.searchInSubfoldersPrefName = "find-searchInSubfolders"
        self.includeFiletypesPrefName = "find-includeFiletypes"
        self.excludeFiletypesPrefName = "find-excludeFiletypes"

        self.patternType = gPrefSvc.prefs.getLongPref(self.patternTypePrefName)
        self.caseSensitivity = gPrefSvc.prefs.getLongPref(self.caseSensitivityPrefName)
        self.matchWord = gPrefSvc.prefs.getBooleanPref(self.matchWordPrefName)
        self.searchBackward = gPrefSvc.prefs.getBooleanPref(self.searchBackwardPrefName)
        self.preferredContextType = gPrefSvc.prefs.getLongPref(self.preferredContextTypePrefName)
        self.displayInFindResults2 = gPrefSvc.prefs.getBooleanPref(self.displayInFindResults2PrefName)

        try:
            self._folders = []
            foldersPref = gPrefSvc.prefs.getPref(self.foldersPrefName)
            for i in range(foldersPref.length):
                self._folders.append(foldersPref.getStringPref(i))
        except COMException, ex:
            log.exception(ex, "Error retrieving 'File in Files' folders "
                              "preference. Some folders may have been lost.")
        self.encodedFolders = os.pathsep.join(self._folders)
        
        self.searchInSubfolders = gPrefSvc.prefs.getBooleanPref(self.searchInSubfoldersPrefName)

        try:
            self._includeFiletypes = []
            includeFiletypesPref = gPrefSvc.prefs.getPref(self.includeFiletypesPrefName)
            for i in range(includeFiletypesPref.length):
                self._includeFiletypes.append(includeFiletypesPref.getStringPref(i))
        except COMException, ex:
            log.exception(ex, "Error retrieving 'File in Files' include "
                              "filetypes preference. Some filetypes may have "
                              "been lost.")
        self.encodedIncludeFiletypes = os.pathsep.join(self._includeFiletypes)
        try:
            self._excludeFiletypes = []
            excludeFiletypesPref = gPrefSvc.prefs.getPref(self.excludeFiletypesPrefName)
            for i in range(excludeFiletypesPref.length):
                self._excludeFiletypes.append(excludeFiletypesPref.getStringPref(i))
        except COMException, ex:
            log.exception(ex, "Error retrieving 'File in Files' exclude "
                              "filetypes preference. Some filetypes may have "
                              "been lost.")
        self.encodedExcludeFiletypes = os.pathsep.join(self._excludeFiletypes)

    def set_patternType(self, value):
        self.patternType = value
        return gPrefSvc.prefs.setLongPref(self.patternTypePrefName, value)

    def set_caseSensitivity(self, value):
        self.caseSensitivity = value
        return gPrefSvc.prefs.setLongPref(self.caseSensitivityPrefName, value)

    def set_matchWord(self, value):
        self.matchWord = value
        return gPrefSvc.prefs.setBooleanPref(self.matchWordPrefName, value)

    def set_searchBackward(self, value):
        self.searchBackward = value
        return gPrefSvc.prefs.setBooleanPref(self.searchBackwardPrefName, value)

    def set_preferredContextType(self, value):
        self.preferredContextType = value
        return gPrefSvc.prefs.setLongPref(self.preferredContextTypePrefName, value)

    def set_displayInFindResults2(self, value):
        self.displayInFindResults2 = value
        return gPrefSvc.prefs.setBooleanPref(self.displayInFindResults2PrefName, value)

    def set_encodedFolders(self, value):
        self.encodedFolders = value
        self._folders = value.split(os.pathsep)
        foldersPref = gPrefSvc.prefs.getPref(self.foldersPrefName)
        foldersPref.reset()
        for folder in self._folders:
            foldersPref.appendStringPref(folder)
        gPrefSvc.prefs.setPref(self.foldersPrefName, foldersPref)

    def getFolders(self):
        return self._folders

    def set_searchInSubfolders(self, value):
        self.searchInSubfolders = value
        return gPrefSvc.prefs.setBooleanPref(self.searchInSubfoldersPrefName, value)

    def set_encodedIncludeFiletypes(self, value):
        self.encodedIncludeFiletypes = value
        self._includeFiletypes = value.split(os.pathsep)
        includeFiletypesPref = gPrefSvc.prefs.getPref(self.includeFiletypesPrefName)
        includeFiletypesPref.reset()
        for includeFiletype in self._includeFiletypes:
            includeFiletypesPref.appendStringPref(includeFiletype)
        gPrefSvc.prefs.setPref(self.includeFiletypesPrefName,
                               includeFiletypesPref)

    def getIncludeFiletypes(self):
        return self._includeFiletypes

    def set_encodedExcludeFiletypes(self, value):
        self.encodedExcludeFiletypes = value
        self._excludeFiletypes = value.split(os.pathsep)
        excludeFiletypesPref = gPrefSvc.prefs.getPref(self.excludeFiletypesPrefName)
        excludeFiletypesPref.reset()
        for excludeFiletype in self._excludeFiletypes:
            excludeFiletypesPref.appendStringPref(excludeFiletype)
        gPrefSvc.prefs.setPref(self.excludeFiletypesPrefName,
                               excludeFiletypesPref)

    def getExcludeFiletypes(self):
        return self._excludeFiletypes


class KoFindService:
    _com_interfaces_ = components.interfaces.koIFindService
    _reg_desc_  = "Find and Replace Service"
    _reg_clsid_ = "{3582DE9B-FA36-4787-B3D3-6B9F94EB4AD0}"
    _reg_contractid_ = "@activestate.com/koFindService;1"

    def __init__(self):
        global gPrefSvc, lastErrorSvc
        gPrefSvc = components.classes["@activestate.com/koPrefService;1"]\
                  .getService(components.interfaces.koIPrefService)
        lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)

        # Define mappings from koIFinder.idl option enums to findlib option
        # values.
        self.patternTypeMap = {    # koIFinder.patternType -> findlib.patternType
            FOT_SIMPLE:      "simple",
            FOT_WILDCARD:    "wildcard",
            FOT_REGEX_PYTHON:"regex-python",
        }
        self.caseMap = {   # koIFinder.caseSensitivity -> findlib.case
            FOC_INSENSITIVE: "insensitive",
            FOC_SENSITIVE:   "sensitive",
            FOC_SMART:       "smart",
        }

        # load the find and replace options
        self.options = KoFindOptions()
        
        self._threadMap = {}
       
    def find(self, url, text, pattern, startOffset):
        """Return the result of searching for the first "pattern" in "text".
        Return value is a KoFindResult or None.
        """
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            result = findlib.find(text, pattern, startOffset=startOffset,
                                  patternType=patternType, case=case,
                                  searchBackward=self.options.searchBackward,
                                  matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if result:
            retval = KoFindResult(url=url, start=result.start, end=result.end,
                                  value=result.value)
        else:
            retval = None
        return retval

    def findall(self, url, text, pattern):
        """Return all the results of searching for "pattern" in "text".
        Return value is a list of KoFindResult's.
        """
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            results = findlib.findall(text, pattern, patternType=patternType,
                                      case=case,
                                      matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        retvals = []
        for result in results:
            retval = KoFindResult(url=url, start=result.start, end=result.end,
                                  value=result.value)
            retvals.append(retval)
        return retvals

    def replace(self, url, text, pattern, replacement, startOffset):
        """Return a result indicating how to replace the first "pattern" in
        "text" with "replacement".

        Return value is a KoReplaceResult or None.
        """
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            result = findlib.replace(text, pattern, replacement,
                                     startOffset=startOffset,
                                     patternType=patternType, case=case,
                                     searchBackward=self.options.searchBackward,
                                     matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if result:
            retval = KoReplaceResult(url=url,
                                     start=result.start, end=result.end,
                                     value=result.value,
                                     replacement=result.replacement)
        else:
            retval = None
        return retval

    def replaceall(self, url, text, pattern, replacement):
        """Return a result indicating how to replace the first "pattern" in
        "text" with "replacement".
        
        DEPRECATED: use replaceallex()

        Return value is a list of KoReplaceResult's.
        """
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            results = findlib.replaceall(text, pattern, replacement,
                                         patternType=patternType,
                                         case=case,
                                         searchBackward=self.options.searchBackward,
                                         matchWord=self.options.matchWord);
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        retvals = []
        for result in results:
            retval = KoReplaceResult(url=url,
                                     start=result.start, end=result.end,
                                     value=result.value,
                                     replacement=result.replacement)
            retvals.append(retval)
        return retvals

    def findallex(self, url, text, pattern, resultsView, contextOffset,
                  scimoz):
        """Feed all occurrences of "pattern" in the "text" into the given
        koIFindResultsView.

            "url" is the viewId.
            "resultsView" is null or a koIFindResultView instance on which
                the replace results should be logged via the Add*() methods.
            "contextOffset" is text's offset into the scimoz buffer. This
                is only used if resultsView is specified.
            "scimoz" is the ISciMoz interface for current view. This is only
                used if resultsView is specified.
        
        No return value.
        """
        log.info("findallex(url='%s', text=%r, pattern='%s', resultsView,"
                 "contextOffset=%s, scimoz)", url, text[:20]+text[-20:],
                 pattern, contextOffset)

        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            matches = findlib.findallex(
                text, pattern,
                patternType=patternType, case=case,
                matchWord=self.options.matchWord)
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if matches:
            resultsView = UnwrapObject(resultsView)
            for match in matches:
                value = match.group()
                startIndex = match.start() + contextOffset
                endIndex = startIndex + len(value)
                startLineNum = scimoz.lineFromPosition(startIndex)
                endLineNum = scimoz.lineFromPosition(endIndex)
                contextStartPos = scimoz.positionFromLine(startLineNum)
                contextEndPos = scimoz.getLineEndPosition(endLineNum)
                context = scimoz.getTextRange(contextStartPos, contextEndPos)
                #context = scimoz.getTextRange(startIndex-10, endIndex+10)
                resultsView.AddFindResult(
                    url, startIndex, endIndex, value,
                    url, # fileName (currently url/viewId is the displayName)
                    startLineNum + 1,
                    startIndex - scimoz.positionFromLine(startLineNum),
                    context)

    def findalllines(self, url, text, pattern, contextOffset, scimoz):
        """Return a lines on which "pattern" is found.

            "url" is the viewId.
            "contextOffset" is text's offset into the scimoz buffer. This
                is only used if resultsView is specified.
            "scimoz" is the ISciMoz interface for current view. This is only
                used if resultsView is specified.
        
        Returns a list of line numbers (0-based).
        """
        
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            matches = findlib.findallex(
                text, pattern,
                patternType=patternType, case=case,
                matchWord=self.options.matchWord)
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        lines = {}  # use dict to avoid dupes
        for match in matches:
            startIndex = match.start() + contextOffset
            startLineNum = scimoz.lineFromPosition(startIndex)
            lines[startLineNum] = 1
        return lines.keys()

    def replaceallex(self, url, text, pattern, replacement, session,
                     resultsView, contextOffset, scimoz):
        """Replace all occurrences of "pattern" in the given text.

            "url" is the viewId.
            "session" is the koIFindSession instance.
            "resultsView" is null or a koIFindResultView instance on which
                the replace results should be logged via the Add*() methods.
            "contextOffset" is text's offset into the scimoz buffer. This
                is only used if resultsView is specified.
            "scimoz" is the ISciMoz interface for current view. This is only
                used if resultsView is specified.
        
        Returns:
            1. the replacement text if there was a change, otherwise null
            2. the number of replacements made
        """
        # Determine a "skip-zone" in the current URL. I.e., the area already
        # covered by the user in this find session.
        session = UnwrapObject(session)
        secondLastHit = session.GetSecondLastFindResult()
        if not secondLastHit or secondLastHit.url != url:
            skipZone = None
        elif session.wrapped:
            # Skip zone is from 0 to end of second last hit and from
            # fileStartPos to end of file.
            if self.options.searchBackward:
                skipZone = [(0, session.fileStartPos),
                            (secondLastHit.start, len(text))]
            else:
                skipZone = [(0, secondLastHit.end),
                            (session.fileStartPos, len(text))]
        else:
            # Skip zone is from the fileStartPos to the end of the second
            # last hit. (_Second_ last, because the current hit (the "last")
            # one is still a candidate for replacement.)
            if self.options.searchBackward:
                skipZone = [(secondLastHit.start, session.fileStartPos)]
            else:
                skipZone = [(session.fileStartPos, secondLastHit.end)]
        
        patternType = self.patternTypeMap[self.options.patternType]
        case = self.caseMap[self.options.caseSensitivity]
        try:
            replacementText, numReplacements, matches = findlib.replaceallex(
                text, pattern, replacement,
                patternType=patternType, case=case,
                searchBackward=self.options.searchBackward,
                matchWord=self.options.matchWord, skipZone=skipZone,
                wantMatches=resultsView is not None)
        except (re.error, findlib.FindError), ex:
            global lastErrorSvc
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        if matches:
            resultsView = UnwrapObject(resultsView)
            for match in matches:
                repl = match.expand(replacement)
                startIndex = match.start() + contextOffset
                endIndex = startIndex + len(repl)
                startLineNum = scimoz.lineFromPosition(startIndex)
                endLineNum = scimoz.lineFromPosition(endIndex)
                contextStartPos = scimoz.positionFromLine(startLineNum)
                contextEndPos = scimoz.getLineEndPosition(endLineNum)
                context = scimoz.getTextRange(contextStartPos, contextEndPos)
                #context = scimoz.getTextRange(startIndex-10, endIndex+10)
                resultsView.AddReplaceResult(
                    url, startIndex, endIndex,
                    match.group(), # value
                    repl, # replacement string
                    url, # fileName (currently url/viewId is the displayName)
                    startLineNum + 1,
                    startIndex - scimoz.positionFromLine(startLineNum),
                    context)

        return replacementText, numReplacements

    def findallinfiles(self, id, pattern, resultsMgr, resultsView):
        """Feed all occurrences of "pattern" in the files identified by
        the options attribute into the given koIFindResultsView.

            "id" is a unique number to distinguish this findallinfiles
                session from others.
            "resultsMgr" is a koIFindResultsTab instance.
            "resultsView" is a koIFindResultView instance on which
                the find results should be logged via the Add*() methods.
        
        This process is done asynchronously -- i.e. a separate thread is
        started to do this.
        
        No return value.
        """
        log.info("findallinfiles(pattern='%s', resultsView)", pattern)

        patternType = self.patternTypeMap[self.options.patternType]
        caseSensitivity = self.caseMap[self.options.caseSensitivity]
        # validate now
        try:
            findlib.validatePattern(pattern, patternType=patternType,
                                    case=caseSensitivity,
                                    matchWord=self.options.matchWord)
        except (re.error, findlib.FindError), ex:
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

        t = _FinderInFiles(id, pattern, patternType, caseSensitivity,
                           self.options.matchWord,
                           self.options.getFolders(),
                           resultsMgr.context_.cwd,
                           self.options.searchInSubfolders,
                           self.options.getIncludeFiletypes(),
                           self.options.getExcludeFiletypes(),
                           resultsMgr,
                           resultsView)
        self._threadMap[id] = t
        resultsMgr.searchStarted()
        self._threadMap[id].start()

    def stopfindallinfiles(self, id):
        #XXX Do I need a lock-guard around usage of self._threadMap?
        if id in self._threadMap:
            self._threadMap[id].stop()
            del self._threadMap[id]


